// define whether RTC is to be used or not
#define USE_RTC 1

// includes for RTC
#if USE_RTC
  #include <Wire.h>
  #include "RTClib.h"
#endif

// include for NeoPixel matrix
#include <Adafruit_NeoPixel.h>

/*
class PixelTimer
which is an object that should make it easy
to programm a large number of neo pixels to
turn on/off at various intervals unrelated
to each other

Florian Richter, Moeglich Lab HU Berlin
florian.richter.1@hu-berlin.de, feb 2014

extended by the incredible AM
andreas.moeglich@uni-bayreuth.de

*/

unsigned int neopixel_pin  = 6;
#define  num_neopixels 64
#define PI 3.14159265358979323846
Adafruit_NeoPixel led_matrix = Adafruit_NeoPixel(num_neopixels, neopixel_pin, NEO_GRB + NEO_KHZ800);

void component_colors_from_uint32t(
  uint32_t color,
  uint8_t & red_val,
  uint8_t & green_val,
  uint8_t & blue_val
 )
 {
  red_val = (uint8_t)(color >> 16),
  green_val = (uint8_t)(color >>  8),
  blue_val = (uint8_t)color;
 }

boolean is_led_set(
  uint8_t which_led, uint8_t set_red, uint8_t set_green, uint8_t set_blue )
 {
 
  uint32_t current_status(  led_matrix.getPixelColor( which_led ) );
  
  uint8_t cur_r, cur_g, cur_b;
  component_colors_from_uint32t( current_status, cur_r, cur_g, cur_b );
  if( (cur_r == set_red ) && (cur_g == set_green ) && (cur_b == set_blue ) ){
    return true;
  }
  else{
    return false;
  }
 }

class PixelTimer {

public:
  PixelTimer();
  
  PixelTimer( 
    uint8_t which_pixel,
    uint8_t intensity_r,
    uint8_t intensity_g,
    uint8_t intensity_b,
    uint8_t intensity_r2,
    uint8_t intensity_g2,
    uint8_t intensity_b2,
    uint8_t mode,
    unsigned long param1, 
    unsigned long param2
  );

  ~PixelTimer();
  
  void set_values(
    uint8_t which_pixel,
    uint8_t intensity_r,
    uint8_t intensity_g,
    uint8_t intensity_b,
    uint8_t intensity_r2,
    uint8_t intensity_g2,
    uint8_t intensity_b2,
    uint8_t mode,
    unsigned long param1, 
    unsigned long param2 
  );
  
  void process_time( unsigned long current_time );
  void init_on();

private:
  uint8_t my_pixel_;
  uint8_t intensity_r_;
  uint8_t intensity_g_;
  uint8_t intensity_b_;
  uint8_t intensity_r2_;
  uint8_t intensity_g2_;
  uint8_t intensity_b2_;
  uint8_t mode_;
  unsigned long param1_, param2_;

  void PixelTimer::boxcar (unsigned long current_time);
  void PixelTimer::sine (unsigned long current_time);
};

PixelTimer::PixelTimer() : my_pixel_(num_neopixels +1), intensity_r_(0), intensity_g_(0), intensity_b_(0), intensity_r2_(0), intensity_g2_(0), intensity_b2_(0), mode_(0), param1_(0), param2_(0)
{}

PixelTimer::PixelTimer(
  uint8_t which_pixel,
  uint8_t intensity_r,
  uint8_t intensity_g,
  uint8_t intensity_b,
  uint8_t intensity_r2,
  uint8_t intensity_g2,
  uint8_t intensity_b2,
  uint8_t mode,
  unsigned long param1,
  unsigned long param2
) : my_pixel_(which_pixel), intensity_r_(intensity_r), intensity_g_(intensity_g), intensity_b_(intensity_b), intensity_r2_(intensity_r2), intensity_g2_(intensity_g2), intensity_b2_(intensity_b2), mode_(mode), param1_(param1), param2_(param2)
{}

PixelTimer::~PixelTimer(){}

void PixelTimer::set_values(
  uint8_t which_pixel,
  uint8_t intensity_r,
  uint8_t intensity_g,
  uint8_t intensity_b,
  uint8_t intensity_r2,
  uint8_t intensity_g2,
  uint8_t intensity_b2,
  uint8_t mode,
  unsigned long param1, 
  unsigned long param2 
)
{
  my_pixel_= which_pixel; 
  intensity_r_ = intensity_r;
  intensity_g_ = intensity_g;
  intensity_b_ = intensity_b;
  intensity_r2_ = intensity_r2;
  intensity_g2_ = intensity_g2;
  intensity_b2_ = intensity_b2;
  mode_ = mode;
  param1_ = param1;
  param2_ = param2;
}

void PixelTimer::init_on()
{
  led_matrix.setPixelColor( my_pixel_, intensity_r_, intensity_g_, intensity_b_);
}


void PixelTimer::process_time(unsigned long current_time )
{
  // switch according to mode
  switch (mode_)
  {
    case 0:
      boxcar(current_time);
      break;
    case 1:
      sine(current_time);
      break;
    default:
      boxcar(current_time);
  }
}

void PixelTimer::boxcar (unsigned long current_time)
{
  unsigned long t_in_current_wave = current_time % (param1_+param2_);

  // check whether we are in on or off phase
  if (t_in_current_wave <= param1_)
  {
    // check LED status
    if (!is_led_set(my_pixel_,intensity_r_, intensity_g_, intensity_b_))
    {
      led_matrix.setPixelColor( my_pixel_, intensity_r_, intensity_g_, intensity_b_);
    }
  }
  else
  {
    // check LED status
    if (!is_led_set(my_pixel_,intensity_r2_, intensity_g2_, intensity_b2_))
    {
      led_matrix.setPixelColor( my_pixel_, intensity_r2_, intensity_g2_, intensity_b2_);
    }
  }
}

void PixelTimer::sine( unsigned long current_time)
{
  unsigned long t_in_current_wave = current_time % param1_;
  double sval;

  sval=pow(sin(2.0*PI*(1.0*t_in_current_wave)/(1.0*param1_)),2);
  led_matrix.setPixelColor( my_pixel_, sval*intensity_r_+(1-sval)*intensity_r2_, sval*intensity_g_+(1-sval)*intensity_g2_, sval*intensity_b_+(1-sval)*intensity_b2_);
}

// global variables
PixelTimer ptimers[num_neopixels];
#if USE_RTC
  RTC_DS3231 rtc;
#endif
// check pixels every how many ms?
int delay_time = 25;
long start_sec;
unsigned long start_millis;
#if USE_RTC
  // specify at which interval (in s) the internal clock is synchronized by the RTC
  long time_correct_interval = 10;
  long prev_sec;
#endif

void setup(){
  // set up NeoPixel matrix
  led_matrix.begin(); // Initialize all pixels  
  // arguments are: pixel ID, red, green, blue, red2, green2, blue2, mode, param1, param2
  ptimers[0].set_values(0, 0, 0, 1, 0, 0, 0, 0, 30000, 1800000);
  ptimers[1].set_values(1, 0, 0, 2, 0, 0, 0, 0, 30000, 1800000);
  ptimers[2].set_values(2, 0, 0, 4, 0, 0, 0, 0, 30000, 1800000);
  ptimers[3].set_values(3, 0, 0, 8, 0, 0, 0, 0, 30000, 1800000);
  ptimers[4].set_values(4, 0, 0, 16, 0, 0, 0, 0, 30000, 1800000);
  ptimers[5].set_values(5, 0, 0, 32, 0, 0, 0, 0, 30000, 1800000);
  ptimers[6].set_values(6, 0, 0, 64, 0, 0, 0, 0, 30000, 1800000);
  ptimers[7].set_values(7, 0, 0, 128, 0, 0, 0, 0, 30000, 1800000);
  ptimers[8].set_values(8, 0, 0, 1, 0, 0, 0, 0, 60000, 1800000);
  ptimers[9].set_values(9, 0, 0, 2, 0, 0, 0, 0, 60000, 1800000);
  ptimers[10].set_values(10, 0, 0, 4, 0, 0, 0, 0, 60000, 1800000);
  ptimers[11].set_values(11, 0, 0, 8, 0, 0, 0, 0, 60000, 1800000);
  ptimers[12].set_values(12, 0, 0, 16, 0, 0, 0, 0, 60000, 1800000);
  ptimers[13].set_values(13, 0, 0, 32, 0, 0, 0, 0, 60000, 1800000);
  ptimers[14].set_values(14, 0, 0, 64, 0, 0, 0, 0, 60000, 1800000);
  ptimers[15].set_values(15, 0, 0, 128, 0, 0, 0, 0, 60000, 1800000);
  ptimers[16].set_values(16, 0, 0, 1, 0, 0, 0, 0, 120000, 1800000);
  ptimers[17].set_values(17, 0, 0, 2, 0, 0, 0, 0, 120000, 1800000);
  ptimers[18].set_values(18, 0, 0, 4, 0, 0, 0, 0, 120000, 1800000);
  ptimers[19].set_values(19, 0, 0, 8, 0, 0, 0, 0, 120000, 1800000);
  ptimers[20].set_values(20, 0, 0, 16, 0, 0, 0, 0, 120000, 1800000);
  ptimers[21].set_values(21, 0, 0, 32, 0, 0, 0, 0, 120000, 1800000);
  ptimers[22].set_values(22, 0, 0, 64, 0, 0, 0, 0, 120000, 1800000);
  ptimers[23].set_values(23, 0, 0, 128, 0, 0, 0, 0, 120000, 1800000);
  ptimers[24].set_values(24, 0, 0, 1, 0, 0, 0, 0, 240000, 1800000);
  ptimers[25].set_values(25, 0, 0, 2, 0, 0, 0, 0, 240000, 1800000);
  ptimers[26].set_values(26, 0, 0, 4, 0, 0, 0, 0, 240000, 1800000);
  ptimers[27].set_values(27, 0, 0, 8, 0, 0, 0, 0, 240000, 1800000);
  ptimers[28].set_values(28, 0, 0, 16, 0, 0, 0, 0, 240000, 1800000);
  ptimers[29].set_values(29, 0, 0, 32, 0, 0, 0, 0, 240000, 1800000);
  ptimers[30].set_values(30, 0, 0, 64, 0, 0, 0, 0, 240000, 1800000);
  ptimers[31].set_values(31, 0, 0, 128, 0, 0, 0, 0, 240000, 1800000);
  ptimers[32].set_values(32, 0, 0, 1, 0, 0, 0, 0, 480000, 1800000);
  ptimers[33].set_values(33, 0, 0, 2, 0, 0, 0, 0, 480000, 1800000);
  ptimers[34].set_values(34, 0, 0, 4, 0, 0, 0, 0, 480000, 1800000);
  ptimers[35].set_values(35, 0, 0, 8, 0, 0, 0, 0, 480000, 1800000);
  ptimers[36].set_values(36, 0, 0, 16, 0, 0, 0, 0, 480000, 1800000);
  ptimers[37].set_values(37, 0, 0, 32, 0, 0, 0, 0, 480000, 1800000);
  ptimers[38].set_values(38, 0, 0, 64, 0, 0, 0, 0, 480000, 1800000);
  ptimers[39].set_values(39, 0, 0, 128, 0, 0, 0, 0, 480000, 1800000);
  ptimers[40].set_values(40, 0, 0, 1, 0, 0, 0, 0, 960000, 1800000);
  ptimers[41].set_values(41, 0, 0, 2, 0, 0, 0, 0, 960000, 1800000);
  ptimers[42].set_values(42, 0, 0, 4, 0, 0, 0, 0, 960000, 1800000);
  ptimers[43].set_values(43, 0, 0, 8, 0, 0, 0, 0, 960000, 1800000);
  ptimers[44].set_values(44, 0, 0, 16, 0, 0, 0, 0, 960000, 1800000);
  ptimers[45].set_values(45, 0, 0, 32, 0, 0, 0, 0, 960000, 1800000);
  ptimers[46].set_values(46, 0, 0, 64, 0, 0, 0, 0, 960000, 1800000);
  ptimers[47].set_values(47, 0, 0, 128, 0, 0, 0, 0, 960000, 1800000);
  ptimers[48].set_values(48, 0, 0, 1, 0, 0, 0, 0, 1800000, 1800000);
  ptimers[49].set_values(49, 0, 0, 2, 0, 0, 0, 0, 1800000, 1800000);
  ptimers[50].set_values(50, 0, 0, 4, 0, 0, 0, 0, 1800000, 1800000);
  ptimers[51].set_values(51, 0, 0, 8, 0, 0, 0, 0, 1800000, 1800000);
  ptimers[52].set_values(52, 0, 0, 16, 0, 0, 0, 0, 1800000, 1800000);
  ptimers[53].set_values(53, 0, 0, 32, 0, 0, 0, 0, 1800000, 1800000);
  ptimers[54].set_values(54, 0, 0, 64, 0, 0, 0, 0, 1800000, 1800000);
  ptimers[55].set_values(55, 0, 0, 128, 0, 0, 0, 0, 1800000, 1800000);
  ptimers[56].set_values(56, 0, 0, 1, 0, 0, 0, 0, 3600000, 1800000);
  ptimers[57].set_values(57, 0, 0, 2, 0, 0, 0, 0, 3600000, 1800000);
  ptimers[58].set_values(58, 0, 0, 4, 0, 0, 0, 0, 3600000, 1800000);
  ptimers[59].set_values(59, 0, 0, 8, 0, 0, 0, 0, 3600000, 1800000);
  ptimers[60].set_values(60, 0, 0, 16, 0, 0, 0, 0, 3600000, 1800000);
  ptimers[61].set_values(61, 0, 0, 32, 0, 0, 0, 0, 3600000, 1800000);
  ptimers[62].set_values(62, 0, 0, 64, 0, 0, 0, 0, 3600000, 1800000);
  ptimers[63].set_values(63, 0, 0, 128, 0, 0, 0, 0, 3600000, 1800000);


#if USE_RTC
  // set up RTC and reset time
  rtc.begin();
  DateTime now = rtc.now();
  start_sec=now.secondstime();
  prev_sec=0;
#endif

  // record start of millis
  start_millis=millis();
} 

void loop() {
   unsigned long curr_millis = millis();
#if USE_RTC
   DateTime now = rtc.now();
   long curr_sec=now.secondstime()-start_sec;

   // synchronize millisecond counter to the RTC
   if (curr_sec-time_correct_interval>=prev_sec)
   {
     prev_sec=curr_sec;
     start_millis=curr_millis;
   }
   // only keep the last millis since last update
   curr_millis+=1000*time_correct_interval*floor(curr_sec/time_correct_interval);
   curr_millis-=start_millis;
#endif

   for( int i = 0; i < num_neopixels; i++){
     ptimers[i].process_time( curr_millis );
   }
   led_matrix.show();
   delay( delay_time );
}